@@ -19,8 +19,11 @@
 
 (define-module (nyacc lang c99 cpp)
   #:export (parse-cpp-stmt
+	    read-cpp-stmt
 	    parse-cpp-expr
 	    eval-cpp-expr
+	    cpp-expand-text
+	    expand-cpp-mref
 	    )
   #:use-module (nyacc parse)
   #:use-module (nyacc lex)
@@ -41,14 +44,14 @@ todo:
   provide util to expand defines
 |#
 
-;;.@item skip-ws ch
+;;.@deffn skip-ws ch
 ;; Helper for 
 (define (skip-ws ch)
   (if (eof-object? ch) ch
       (if (char-set-contains? c:ws ch)
 	  (skip-ws (read-char))
 	  ch)))
-;; @item cpp-define => #f|???
+;; @deffn cpp-define => #f|???
 (define (cpp-define)
   ;; The (weak?) parse architecture is "unread la argument if no match"
   (letrec
@@ -58,10 +61,9 @@ todo:
 		 ;;(args (or (p-args (skip-ws (read-char))) '()))
 		 ;; "define ABC(ARG)" not the same as "define ABC (ARG)"
 		 (args (or (p-args (read-char)) '()))
-		 (amap (lambda (as) (map (lambda (a) (list 'arg a)) as)))
 		 (rest (or (p-rest (skip-ws (read-char))) " ")))
 	    (if (pair? args)
-		`(define (name ,iden) ,(cons 'args (amap args)) (repl ,rest))
+		`(define (name ,iden) ,(cons 'args args) (repl ,rest))
 		`(define (name ,iden) (repl ,rest))))))
        (p-args ;; parse args
 	(lambda (la) ;; unread la if no match :(
@@ -80,7 +82,7 @@ todo:
 		(else #f)))))
     (p-cppd)))
 
-;; @item cpp-include
+;; @deffn cpp-include
 ;; Parse CPP include statement.
 (define (cpp-include)
   (let* ((beg-ch (skip-ws (read-char)))
@@ -90,39 +92,34 @@ todo:
 		     (iter (cons ch cl) (read-char))))))
     `(include ,path)))
 
-;; @item parse-cpp-stmt line => tree
+;; @deffn read-cpp-stmt line defs => (stmt-type text)
 ;; Parse a line from a CPP statement and return a parse tree.
 ;; @example
 ;; (parse-cpp-stmt "define X 123") => (define "X" "123")
 ;; (parse-cpp-stmt "if defined(A) && defined(B) && defined(C)"
-;; => (if (and (and (defined "A") (defined "B"))
-;;             (not (defined "C")))
+;; => (if "defined(A) && defined(B) && defined(C)")
 ;; @end example
-(define (parse-cpp-stmt line)
+;; To evaluate the @code{if} statements use @code{parse-cpp-expr} and
+;; @code{eval-cpp-expr}.
+(define (read-cpp-stmt line)
+  (define (rd-ident) (read-c-ident (skip-ws (read-char))))
+  (define (rd-num) (and=> (read-c-num (skip-ws (read-char))) cdr))
+  (define (rd-rest) (let ((ch (skip-ws (read-char))))
+		      (if (not (eof-object? ch)) (unread-char ch))
+		      (drain-input (current-input-port))))
   (with-input-from-string line
     (lambda ()
-      (let ((cmd (string->symbol (read-c-ident (skip-ws (read-char)))))
-	    (rd-ident
-	     #;(lambda () (or (and=> (read-c-ident (skip-ws (read-char))) cdr)
-	     (throw 'parse-error)))
-	     (lambda () (read-c-ident (skip-ws (read-char))))
-	     )
-	    (rd-num
-	     (lambda () (or (and=> (read-c-num (skip-ws (read-char))) cdr)
-			    (throw 'parse-error))))
-	    )
+      (let ((cmd (string->symbol (read-c-ident (skip-ws (read-char))))))
 	 (case cmd
 	   ((include) (cpp-include))
-	   ((ifdef) `(if (defined ,(rd-ident))))
-	   ((ifndef) `(if (not (defined ,(rd-ident)))))
 	   ((define) (cpp-define))
-	   ((if elif) (list cmd (parse-cpp-expr)))
-	   ((else endif) (list cmd))
 	   ((undef) `(undef ,(rd-ident)))
-	   ((line) `(line ,(rd-num)))
-	   ((error) (read-char) `(error ,(drain-input (current-input-port))))
-	   ;;((pragma) (cpp-define)) ; ???
-	   (else '()))))))
+	   ((ifdef)
+	    `(if ,(string-append "defined(" (rd-ident) ")" (rd-rest))))
+	   ((ifndef)
+	    `(if ,(string-append "!defined(" (rd-ident) ")" (rd-rest))))
+	   ((if elif else endif line error pragma) (list cmd (rd-rest)))
+	   (else '(unknown "")))))))
     
 (include-from-path "nyacc/lang/c99/mach.d/cpptab.scm")
 (include-from-path "nyacc/lang/c99/mach.d/cppact.scm")
@@ -132,8 +129,10 @@ todo:
    (list (cons 'len-v len-v) (cons 'pat-v pat-v) (cons 'rto-v rto-v)
 	 (cons 'mtab mtab) (cons 'act-v act-v))))
 
-;; cppbody provides: gen-cpp-lexer parse-cpp-expr eval-cpp-expr
+;; The included file "cppbody.scm" provides:
+;; gen-cpp-lexer
+;; parse-cpp-expr
+;; eval-cpp-expr
 (include-from-path "nyacc/lang/c99/cppbody.scm")
  
-;; mode: scheme ***
 ;; --- last line ---
